@@ -1,7 +1,7 @@
 /*
  * cowdancer -- a Copy-on-write data-access; No-cow-easy-replacement
  *
- * Copyright 2005-2007 Junichi Uekawa
+ * Copyright 2005-2009 Junichi Uekawa
  * GPL v2 or later.
  */
 #define _GNU_SOURCE
@@ -35,10 +35,26 @@ static int (*origlibc_fchown)(int fd, uid_t, gid_t) = NULL;
 static int (*origlibc_lchown)(const char *, uid_t, gid_t) = NULL;
 static int (*origlibc_chmod)(const char *, mode_t) = NULL;
 static int (*origlibc_fchmod)(int fd, mode_t) = NULL;
+static int (*origlibc_flock)(int fd, int) = NULL;
 
 static struct ilist_struct* ilist=NULL;
 static long ilist_len=0;
 
+/*
+   verify the header content
+ */
+int verify_ilist_header(struct ilist_header header)
+{
+  if (header.revision != ILISTREVISION ||
+      header.ilistsig != ILISTSIG ||
+      header.ilist_struct_size != sizeof (struct ilist_struct))
+    {
+      ilist_outofmemory(".ilist header unexpected");
+      return 1;
+    }
+  return 0;
+}
+
 /* load ilist file
 
    @return 1 on error, 0 on success */
@@ -48,14 +64,15 @@ static int load_ilist(void)
   int fd=0;
   struct stat stbuf;
   struct ilist_struct* local_ilist=NULL;
-  long local_ilist_len=0;  
+  struct ilist_header header;
+  long local_ilist_len=0;
 
   if (!getenv("COWDANCER_ILISTFILE"))
     {
       fprintf(stderr, "env var COWDANCER_ILISTFILE not defined\n");
       return 1;
     }
-  
+
   if (-1==(fd=origlibc_open(getenv("COWDANCER_ILISTFILE"),O_RDONLY,0)))
     {
       fprintf(stderr, "%s: cannot open ilistfile %s\n", ilist_PRGNAME, getenv("COWDANCER_ILISTFILE"));
@@ -67,16 +84,17 @@ static int load_ilist(void)
       return 1;
     }
 
-  local_ilist_len=stbuf.st_size / sizeof(struct ilist_struct);
+  local_ilist_len=(stbuf.st_size - sizeof (struct ilist_header)) / sizeof(struct ilist_struct);
 
-  if (stbuf.st_size != (sizeof(struct ilist_struct) * local_ilist_len))
+  if (stbuf.st_size != (sizeof(struct ilist_struct) * local_ilist_len + sizeof (struct ilist_header)))
     {
+      fprintf(stderr, "%s: .ilist size: %li\n", ilist_PRGNAME, (long)stbuf.st_size);
       ilist_outofmemory(".ilist size unexpected");
       return 1;
     }
-  
+
   if (((void*)-1)==
-      (local_ilist=mmap(NULL, stbuf.st_size, PROT_READ, MAP_PRIVATE, 
+      (local_ilist=mmap(NULL, stbuf.st_size, PROT_READ, MAP_PRIVATE,
 		  fd, 0)))
     {
       perror("mmap failed, failback to other method");
@@ -86,19 +104,25 @@ static int load_ilist(void)
 	  fprintf(stderr, "%s: cannot fdopen ilistfile %s\n", ilist_PRGNAME, getenv("COWDANCER_ILISTFILE"));
 	  return 1;
 	}
-      
+
       if (NULL==(local_ilist=malloc(stbuf.st_size)))
 	{
 	  fprintf(stderr, "%s: out of memory while trying to allocate memory for ilist\n", ilist_PRGNAME);
 	  return 1;
 	}
+      fread(&header, sizeof(struct ilist_header), 1, f);
+      if(verify_ilist_header(header)) return 1;
       fread(local_ilist, sizeof(struct ilist_struct), local_ilist_len, f);
       fclose(f);
     }
-  else 
-    close(fd);
+  else
+    {
+      if (verify_ilist_header(*(struct ilist_header *)local_ilist)) return 1;
+      local_ilist=(void*)local_ilist+sizeof (struct ilist_header);
+      close(fd);
+    }
 
-  /* commit, this should really be atomic, 
+  /* commit, this should really be atomic,
      but I don't want to have a lock here.
      Pray that same result will result in multiple invocations,
      and ask for luck. It shouldn't change the result very much if it was called multiple times.
@@ -108,13 +132,13 @@ static int load_ilist(void)
      at this exact timing, to make this quasi-lock-free logic to be remotely successful. */
   ilist_len = local_ilist_len;
   ilist = local_ilist;
-  
+
   return 0;
 }
 
 static void debug_cowdancer (const char * s)
 {
-  if (getenv("COWDANCER_DEBUG")) 
+  if (getenv("COWDANCER_DEBUG"))
     fprintf (stderr, "%s: DEBUG %s\n", ilist_PRGNAME, s);
 }
 
@@ -124,19 +148,23 @@ static void debug_cowdancer_2 (const char * s, const char*e)
     fprintf (stderr, "%s: DEBUG %s:%s\n", ilist_PRGNAME, s, e);
 }
 
-/** 
+/**
     initialize functions to override libc functions.
 
     @return 1 on error
  */
-__attribute__ ((warn_unused_result)) 
+__attribute__ ((warn_unused_result))
 static int initialize_functions ()
 {
   static volatile int initialized = 0;
-  /* this code is quasi-reentrant; 
+  /* this code is quasi-reentrant;
      shouldn't suffer too much if it is called multiple times. */
-  
-  if (!initialized)
+
+  /* this code is __unlikely__ to be true
+     It is initialized==0 only for the first time
+     so should be !initialized==0 most of the time.
+  */
+  if (__builtin_expect(!initialized,0))
     {
       initialized = 1;
       origlibc_open = dlsym(RTLD_NEXT, "open");
@@ -161,24 +189,26 @@ static int initialize_functions ()
 	    {
 	      debug_cowdancer(d);
 	    }
-	  
-	  
+
+
 	  /* fallback to loading unversioned symbol doing it anyway
 	     since glibc does not seem to set dlerror on dlsym failure.
 	  */
 	  origlibc_chown = dlsym(RTLD_NEXT, "chown");
-	  
+
 	}
       origlibc_fchown = dlsym(RTLD_NEXT, "fchown");
       origlibc_lchown = dlsym(RTLD_NEXT, "lchown");
       origlibc_chmod = dlsym(RTLD_NEXT, "chmod");
       origlibc_fchmod = dlsym(RTLD_NEXT, "fchmod");
+      origlibc_flock = dlsym(RTLD_NEXT, "flock");
 
       if (getenv("COWDANCER_DEBUG"))
 	{
-	  printf("chown:%p lchown:%p\n", origlibc_chown, origlibc_lchown);
+	  fprintf(stderr,
+		  "chown:%p lchown:%p\n", origlibc_chown, origlibc_lchown);
 	}
-      
+
       /* load the ilist */
       if (!ilist)
 	{
@@ -190,11 +220,14 @@ static int initialize_functions ()
 	  initialized = 2;
 	  debug_cowdancer ("Initialization successfully finished.\n");
 	}
-    }  
-  /* wait until somebody else finishes his job */
-  while (initialized == 1)
+    }
+  /*
+     Wait until somebody else finishes his job
+     This is very unlikely
+  */
+  while (__builtin_expect(initialized == 1,0))
     sched_yield();
-  
+
   if (initialized==0)
     return 1;			/* that somebody else failed */
   else
@@ -202,22 +235,26 @@ static int initialize_functions ()
 }
 
 /* the constructor for this library */
-__attribute__ ((constructor)) 
+__attribute__ ((constructor))
      void ctor()
 {
-  initialize_functions();
+  if (initialize_functions())
+    {
+      fprintf(stderr,
+	      "%s: Fatal, initialize_functions failed\n", ilist_PRGNAME);
+    }
 }
 
 /**
    check if i-node is to be protected, and if so, copy the file.  This
    function may fail, but the error cannot really be recovered; how
    should the default be ?
-   
+
    canonicalize flag should be 1 except for non-symlink-following functions.
 
    @return 1 on failure, 0 on success
 */
-__attribute__ ((warn_unused_result)) 
+__attribute__ ((warn_unused_result))
 static int check_inode_and_copy(const char* s, int canonicalize)
 {
   struct ilist_struct search_target;
@@ -227,19 +264,20 @@ static int check_inode_and_copy(const char* s, int canonicalize)
   int ret;
   pid_t pid;
   int status;
-  
+  sigset_t newmask, omask;
+
   debug_cowdancer_2("DEBUG: test for", s);
   if(lstat(s, &buf))
-    return 0;			/* if stat fails, the file probably 
+    return 0;			/* if stat fails, the file probably
 				   doesn't exist; return, success */
   if (canonicalize && S_ISLNK(buf.st_mode))
     {
       /* for symbollic link, canonicalize and get the real filename */
       if (!(canonical=canonicalize_file_name(s)))
-	return 0;			/* if canonicalize_file_name fails, 
+	return 0;			/* if canonicalize_file_name fails,
 					   the file probably doesn't exist. */
-      
-      if(stat(canonical, &buf))	/* if I can't stat this file, I don't think 
+
+      if(stat(canonical, &buf))	/* if I can't stat this file, I don't think
 				   I can write to it; ignore */
 	return 0;
     }
@@ -253,20 +291,21 @@ static int check_inode_and_copy(const char* s, int canonicalize)
   //do some hardcore debugging here:
   if (getenv("COWDANCER_DEBUG"))
     {
-      printf ("ciac: s:%s=canonical:%s nlink:%i reg:%i lnk:%i match:%p\n", s, canonical, (int)buf.st_nlink, S_ISREG(buf.st_mode), S_ISLNK(buf.st_mode),
-	      bsearch(&search_target, ilist, ilist_len, 
+      fprintf (stderr,
+	       "ciac: s:%s=canonical:%s nlink:%i reg:%i lnk:%i match:%p\n", s, canonical, (int)buf.st_nlink, S_ISREG(buf.st_mode), S_ISLNK(buf.st_mode),
+	      bsearch(&search_target, ilist, ilist_len,
 		      sizeof(search_target), compare_ilist));
     }
-  
+
   if((buf.st_nlink > 1) && 	/* it is hardlinked */
-     (S_ISREG(buf.st_mode) || 
+     (S_ISREG(buf.st_mode) ||
       S_ISLNK(buf.st_mode)) && 	/* it is a regular file or a symbollic link */
-     bsearch(&search_target, ilist, ilist_len, 
+     bsearch(&search_target, ilist, ilist_len,
 	     sizeof(search_target), compare_ilist)) /* and a match is
 						       found in ilist
 						       file */
     {
-      /* There is a file that needs to be protected, 
+      /* There is a file that needs to be protected,
 	 Copy-on-write hardlink files.
 
 	 we copy the file first to a backup place, and then mv the
@@ -278,59 +317,87 @@ static int check_inode_and_copy(const char* s, int canonicalize)
 	  ilist_outofmemory("out of memory in check_inode_and_copy, 1");
 	  goto error_canonical;
 	}
-      
+
       close(ret=mkstemp(backup_file));
       if (ret==-1)
 	{
 	  perror("mkstemp");
 	  goto error_buf;
 	}
-      
-      /* let cp do the task, 
-	 it probably knows about filesystem details more than I do. */
+
+      /* mask the SIGCHLD signal, I want waitpid to work.
+       */
+      sigemptyset (&newmask);
+      sigaddset (&newmask, SIGCHLD);
+      if (sigprocmask (SIG_BLOCK, &newmask, &omask) < 0)
+	{
+	  perror("sigprocmask");
+	  goto error_buf;
+	}
+
+      /* let cp do the task,
+	 it probably knows about filesystem details more than I do.
+
+	 forking here is really difficult; I should rewrite this code
+	 to not fork/exec. Signal handling and process waiting is too
+	 unreliable.
+      */
       switch(pid=fork())
 	{
 	case 0:
 	  /* child process, run cp */
 	  putenv("COWDANCER_IGNORE=yes");
+	  sigprocmask (SIG_SETMASK, &omask, (sigset_t *) NULL); /* unmask SIGCHLD signal */
 	  execl("/bin/cp", "/bin/cp", "-a", canonical, backup_file, NULL);
+	  perror("execl:cp:");
 	  exit(EXIT_FAILURE);
 	case -1:
 	  /* error condition in fork(); something is really wrong */
 	  ilist_outofmemory("out of memory in check_inode_and_copy, 2");
-	  goto error_buf;
+	  goto error_spm;
 	default:
 	  /* parent process, waiting for cp -a to terminate */
-	  waitpid(pid, &status, 0);
+	  if(-1==waitpid(pid, &status, 0))
+	    {
+	      perror("waitpid:cp");
+	      fprintf(stderr, "%s: unexpected waitpid error when waiting for process %i with status %x\n",
+		      ilist_PRGNAME, pid, status);
+	      goto error_spm;
+	    }
+
 	  if (!WIFEXITED(status))
 	    {
 	      /* something unexpected */
-	      ilist_outofmemory("unexpected WIFEXITED status in waitpid");
-	      goto error_buf;
+	      fprintf(stderr, "%s: unexpected WIFEXITED status in waitpid: %x\n", ilist_PRGNAME,
+		      (int)status);
+	      goto error_spm;
 	    }
 	  else if (WEXITSTATUS(status))
 	    {
 	      /* cp -a failed */
 	      fprintf(stderr, "%s: cp -a failed for %s\n", ilist_PRGNAME, backup_file);
-	      goto error_buf;
+	      goto error_spm;
 	    }
 	  /* when cp -a succeeded, overwrite the target file from the temporary file with rename */
 	  else if (-1==rename(backup_file, canonical))
 	    {
 	      perror ("file overwrite with rename");
 	      fprintf(stderr, "%s: while trying rename of %s to %s\n",  ilist_PRGNAME, canonical, backup_file);
-	      goto error_buf;
+	      goto error_spm;
 	    }
 	}
       free(backup_file);
+      sigprocmask (SIG_SETMASK, &omask, (sigset_t *) NULL);
     }
-  else				
+  else
     debug_cowdancer_2("DEBUG: did not match ", s);
-  
+
   free(canonical);
   return 0;
 
   /* error-processing routine. */
+ error_spm:
+  sigprocmask (SIG_SETMASK, &omask, (sigset_t *) NULL);
  error_buf:
   free(backup_file);
  error_canonical:
@@ -351,7 +418,7 @@ int open(const char * a, int flags, ...)
       errno=ENOMEM;
       return -1;
     }
-  
+
   if(!getenv("COWDANCER_IGNORE"))
     {
       debug_cowdancer_2 ("open", a);
@@ -527,7 +594,7 @@ int chown(const char* s, uid_t u, gid_t g)
  * @return 1 on failure, 0 on success
  *
  */
-int check_fd_inode_and_warn(int fd)
+int check_fd_inode_and_warn(int fd, const char* operation)
 {
   struct stat buf;
   struct ilist_struct search_target;
@@ -536,20 +603,21 @@ int check_fd_inode_and_warn(int fd)
   memset(&search_target, 0, sizeof(search_target));
   search_target.inode = buf.st_ino;
   search_target.dev = buf.st_dev;
-  if(bsearch(&search_target, ilist, ilist_len, 
+  if(bsearch(&search_target, ilist, ilist_len,
 	     sizeof(search_target), compare_ilist) &&
      S_ISREG(buf.st_mode))
     {
       /* Someone opened file read-only, and called
-	 fchown/fchmod; I don't really know how to do
-	 salvation in that case, since the original filename is 
+	 fchown/fchmod/flock; I don't really know how to do
+	 salvation in that case, since the original filename is
 	 probably not available, and file is already open.
 
 	 If there is any better way, I'd like to know.
        */
-      fprintf(stderr, "Warning: cowdancer: unsupported operation, read-only open and fchown/fchmod: %li:%li\n", 
-	      (long)buf.st_dev, (long)buf.st_ino);
-      /* emit a warning and do not fail, 
+      fprintf(stderr,
+	      "W: cowdancer: unsupported operation %s, read-only open and fchown/fchmod/flock are not supported: tried opening dev:inode of %li:%li\n",
+	      operation, (long)buf.st_dev, (long)buf.st_ino);
+      /* emit a warning and do not fail,
 	 if you want to make it fail, add a return 1;
 	 apt seems to want to use this operation; thus apt will always fail.
        */
@@ -566,7 +634,7 @@ int fchown(int fd, uid_t u, gid_t g)
   if(!getenv("COWDANCER_IGNORE"))
     {
       debug_cowdancer ("fchown");
-      if (check_fd_inode_and_warn(fd))
+      if (check_fd_inode_and_warn(fd, "fchown"))
 	{
 	  errno=ENOMEM;
 	  return -1;
@@ -634,7 +702,7 @@ int fchmod(int fd, mode_t mode)
   if(!getenv("COWDANCER_IGNORE"))
     {
       debug_cowdancer ("fchmod");
-      if (check_fd_inode_and_warn(fd))
+      if (check_fd_inode_and_warn(fd, "fchmod"))
 	{
 	  errno=ENOMEM;
 	  return -1;
@@ -643,3 +711,25 @@ int fchmod(int fd, mode_t mode)
   ret = origlibc_fchmod(fd, mode);
   return ret;
 }
+
+#undef flock
+int flock(int fd, int operation)
+{
+  int ret;
+  if(initialize_functions())
+    {
+      errno=ENOMEM;
+      return -1;
+    }
+  if(!getenv("COWDANCER_IGNORE"))
+    {
+      debug_cowdancer ("flock");
+      if (check_fd_inode_and_warn(fd, "flock"))
+	{
+	  errno=ENOMEM;
+	  return -1;
+	}
+    }
+  ret = origlibc_flock(fd, operation);
+  return ret;
+}
