@@ -161,6 +161,10 @@ int load_config_file(const char* config, pbuilderconfig* pc)
 	    {
 	      pc->arch_diskdevice=strdup(delim);
 	    }
+	  else if (!strcmp(buf, "HTTP_PROXY"))
+	    {
+	      pc->http_proxy=strdup(delim);
+	    }
 	}
     }
 
@@ -206,6 +210,7 @@ int cpbuilder_dumpconfig(pbuilderconfig* pc)
   DUMPSTR(othermirror);
   DUMPSTR(debbuildopts);
   DUMPINT(binary_arch);
+  DUMPSTR(http_proxy);
   DUMPSTRARRAY(inputfile);
   DUMPSTRARRAY(outputfile);
 
@@ -259,6 +264,7 @@ int parse_parameter(int ac, char** av,
     {"inputfile", required_argument, 0, 0},
     {"outputfile", required_argument, 0, 0},
     {"architecture", required_argument, 0, 0},
+    {"http-proxy", required_argument, 0, 0},
 
     /* cowbuilder specific options */
     {"no-cowdancer-update", no_argument, 0, 0},
@@ -268,7 +274,6 @@ int parse_parameter(int ac, char** av,
     {"arch-diskdevice", no_argument, 0, 0},
 
     /* verbatim options with argument, synced as of pbuilder 0.153 */
-    {"http-proxy", required_argument, 0, 'M'},
     {"aptcache", required_argument, 0, 'M'},
     {"extrapackages", required_argument, 0, 'M'},
     {"hookdir", required_argument, 0, 'M'},
@@ -454,6 +459,14 @@ int parse_parameter(int ac, char** av,
 	      /* this is for qemubuilder */
 	      pc.othermirror=strdup(optarg);
 
+	      /* pass it for cowbuilder */
+	      PASS_TO_PBUILDER_WITH_PARAM
+	    }
+	  else if (!strcmp(long_options[index_point].name,"http-proxy"))
+	    {
+	      /* this is for qemubuilder */
+	      pc.http_proxy=strdup(optarg);
+
 	      /* pass it for cowbuilder */
 	      PASS_TO_PBUILDER_WITH_PARAM
 	    }
