@@ -37,7 +37,10 @@ typedef struct pbuilderconfig
   char* mirror;
   char* distribution;
   char* components;
+  char* othermirror;
   char* debbuildopts;
+  int binary_arch;
+  char* http_proxy;
 
   /* files to be copied into the chroot, 
      and copied out of the chroot */
